#include <cassert>

#include "machine.hpp"

// A class that can be have only a certain max number
// of objects at any time (if you attemp to create more it throws)
class instance_counter
{
    enum { max=100 };
    static int count;

    public:
    instance_counter() { if(++count > max) throw "Stack Overflow"; }

    ~instance_counter() { --count; }
};

int instance_counter::count=0;

std::list<function>::const_iterator machine::fetch_function(char name) const
{
	for(std::list<function>::const_iterator it=m_pgm.begin(); it!=m_pgm.end(); ++it){
		if(name==it->name) return it;
	}
	return m_pgm.end();
}

std::list<machine::arg_value>::const_iterator machine::fetch_arg_value(char name) const
{
	for(std::list<arg_value>::const_iterator it=cur_frame().args.begin();
			it!=cur_frame().args.end(); ++it){
		if(name==it->name) return it;
	}
	return cur_frame().args.end();
}

int machine::eval(exp const& e) const
{
	int ret=0;
	for(exp::const_iterator it=e.begin(); it!=e.end(); ++it){
		if(it->type==term::number_type){
			ret+=it->number*(it->neg?-1:1);
		} else {
			std::list<arg_value>::const_iterator itt=fetch_arg_value(it->arg);
			if(itt==cur_frame().args.end()) throw "Argument not found";
			else if(itt->type==arg_value::ins_type) throw "Expected number value.";
			else {
				ret+=itt->number*(it->neg?-1:1);
			}
		}
	}
	return ret;
}

void machine::push_frame(frame const& f)
{
	if(m_exec_stack.size()>500) throw "Stack Overflow";
	m_exec_stack.push_back(f);
}

void machine::load(std::list<function> const& pg, std::list<machine::arg_value> const& arg)
{
	m_pgm=pg;

	if(! m_exec_stack.empty())
		m_exec_stack.clear();

	// Create a new frame whose IP is the first instruction of the
	// last function
	frame f;
	f.ip.fun=m_pgm.end(); f.ip.fun--;
	f.ip.ins=m_pgm.back().stats.begin();
	f.args=arg;

	// Becomes the current (and only) exec frame
	push_frame(f);
}

bool machine::exec(std::vector<char>& res)
{
    // Prevent infinit recursion
    instance_counter cnt;

    if(m_exec_stack.empty()) return false;

	if(cur_frame().ip.ins==cur_frame().ip.fun->stats.end()){
		// No more instructions in the current function?
		// return to the caller!
		bool const ret=this->pop_frame();
		if(ret==false){
			// stack frame empty? program finished!
			return false;
		} else return true;
	}

	// extract current instruction
	stem const st=*cur_frame().ip.ins;

	// Move IP to the next instruction
	cur_frame().ip.ins++;

	if(st.type==stem::s) res.push_back('s');
	else if(st.type==stem::l) res.push_back('l');
	else if(st.type==stem::r) res.push_back('r');
	else if(st.type==stem::arg_type){
		std::list<arg_value>::const_iterator it=fetch_arg_value(st.name);
		if(it==cur_frame().args.end()){
		    --cur_frame().ip.ins;
		    throw "Unresolved id";
		}
		if(it->type==arg_value::number_type) {
            --cur_frame().ip.ins;
		    throw "Expected instruction param";
		}
		res.insert(res.end(), it->ins.begin(), it->ins.end());
	} else {
		// A function call
		assert(st.type==stem::fun_call_type);

		std::list<function>::const_iterator fit=fetch_function(st.name);
		if(fit==m_pgm.end()){
		    --cur_frame().ip.ins;
            throw "Unresolved function name";
		}
		if(st.params.size()!=fit->argument.size()) {
            --cur_frame().ip.ins;
		    throw "Wrong number of argument";
		}

		std::list<param>::const_iterator pit=st.params.begin();
		std::list<char>::const_iterator ait=fit->argument.begin();

		frame fr; // Bind arg names to values
		for(; pit!= st.params.end(); ++pit, ++ait){
			arg_value a;
			a.name=*ait;
			if(pit->type==param::exp_type){
				int val=eval(pit->exp);
				a.type=arg_value::number_type;
				a.number=val;
				if(val<=0) return true;
			} else {
			    // A param like 'B' could be either a numeric param, or
			    // a statement param depending of what B stands for.
			    // The parser, however, always reports it as a statement
			    // param, thus we need to check it here
			    if(pit->stats.size()==1 &&
                    pit->stats.front().type==stem::arg_type) {
                    // This param effetively consists of one 'A' (or 'B' ..)
                    std::list<arg_value>::const_iterator cit=
                        fetch_arg_value(pit->stats.front().name);
                    if(cit!=cur_frame().args.end() &&
                        cit->type==arg_value::number_type){
                        a.type=arg_value::number_type;
                        a.number=cit->number;
                        fr.args.push_back(a);
                        continue;
                    }
                }

				// This param is statement_type need to recursively execute it
				std::list<function> pgm(m_pgm);
				function fun;fun.name='*'; // Unusable from programs
				fun.argument=cur_frame().ip.fun->argument;
				fun.stats=pit->stats;

				pgm.push_back(fun);

				// A machine inside the machine to execute statement params
				machine cpu;
				cpu.load(pgm, cur_frame().args);

				std::vector<char> buffer;
				while(cpu.exec(a.ins));
				a.type=arg_value::ins_type;
			}
			fr.args.push_back(a);
		}
		fr.ip.fun=fit;
		fr.ip.ins=fit->stats.begin();
		if(cur_frame().ip.ins!=cur_frame().ip.fun->stats.end())
            push_frame(fr);
        else cur_frame()=fr;
	}
	return true;
}
