#include <algorithm>
#include <deque>
#include <string>
#include <cctype>

#define _WIN32_IE 0xffff

#include <windows.h>
#include <commctrl.h>

#include "conf.hpp"
#include "gfx.hpp"
#include "resource.hpp"
#include "level.hpp"
#include "global.hpp"

#include "data.hpp"
#include "parser.hpp"
#include "machine.hpp"
#include "about_box.hpp"
#include "bug_box.hpp"

LRESULT CALLBACK main_wnd_proc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK playground_wnd_proc(HWND, UINT, WPARAM, LPARAM);

// The virtual machine executing programs (generating instructions)
machine* cpu;

// The output buffer used to store pending instuctions generated by cpu
// and that still await to be executed by the robot
std::deque<char> instruction_buffer;

// The zoom factor of the playground (1 for 100%, 2 for 200%,.. etc)
int zoom_factor=1;

// Used to temporarily hold program text, and as a general purpuse char buffer
char program_buffer[1000];

// Holds the filename of the currently opened level
// e.g "c:\hello\world\mylevel.txt"
std::string level_filename;

// Pointer inside program_buffer to the current char during
// analysis
char const* program_buffer_it;

// Id used to the Timer pacing the execution
int const timer_id=100;

// Interval used for normal speed
int const timer_interv_normal=70;

// Interval used for fast speed
int const timer_interv_fast=20;

// declared in global.hpp
HINSTANCE global::hinstance;

// The edit where code is entered
HWND edit;

// The toolbar containing the play/stop.. buttons
HWND toolbar;

// The status bar
HWND status;

// The progress bar showing score
HWND score_progress;

// The label showing score
HWND score_label;

// The playground (showing the level&herbert..)
HWND playground;

// The currently loaded and shown level
level current_level;

// Dynamic info about the current level
dyn_level current_dyn_level;

namespace state{
enum type { running, stopped, paused };
}

namespace speed {
enum type {step, normal, fast };
}

// The current state of the application
state::type current_state=state::stopped;

// The currently selected speed
speed::type current_speed=speed::normal;

// Invlidate the playground area
void invalidate_playground(HWND hwnd)
{
    RECT dirt;
    dirt.left=0; dirt.top=0;
    dirt.right=zoom_factor*conf::playground_width;
    dirt.bottom=zoom_factor*conf::playground_height;
    InvalidateRect(GetDlgItem(hwnd, IDC_PLAYGROUND), &dirt, false);
}

// Start running the program
void start_running(HWND hwnd)
{
    if(current_state==state::running) return;

    switch(current_speed){
        case speed::step:
        case speed::normal:
            SetTimer(hwnd, timer_id, timer_interv_normal, NULL);
            break;
        case speed::fast:
            SetTimer(hwnd, timer_id, timer_interv_fast, NULL);
        default: break;
    }
    SetWindowText(hwnd, "Herbert (running)");
    current_state=state::running;

    // Turn the play button into a pause button
    TBBUTTONINFO tbbi;
    tbbi.cbSize=sizeof(tbbi);
    tbbi.dwMask=TBIF_COMMAND | TBIF_IMAGE | TBIF_TEXT ;
    tbbi.iImage=2;
    tbbi.idCommand=IDM_PAUSE;
    tbbi.pszText="Pause";
    tbbi.cchText=sizeof("Pause");

    SendMessage(toolbar, TB_SETBUTTONINFO, IDM_PLAY, (LPARAM)&tbbi);

    // Set the edit area read only
    SendMessage(edit, EM_SETREADONLY, true, 0);
}

// Pause running the program
void pause_running(HWND hwnd)
{
    if(current_state==state::paused) return;

    KillTimer(hwnd, timer_id);
    SetWindowText(hwnd, "Herbert (paused)");

    // Turn the play button into a pause button
    TBBUTTONINFO tbbi;
    tbbi.cbSize=sizeof(tbbi);
    tbbi.dwMask=TBIF_COMMAND | TBIF_IMAGE | TBIF_TEXT ;
    tbbi.iImage=0;
    tbbi.idCommand=IDM_PLAY;
    tbbi.pszText="Play";
    tbbi.cchText=sizeof("Play");

    SendMessage(toolbar, TB_SETBUTTONINFO, IDM_PAUSE, (LPARAM)&tbbi);
    current_state=state::paused;
}


// Saves current sollution to appropriate file
// return true on success, false on error
bool autosave(HWND hwnd)
{
    if(level_filename.empty()) return true;

    FILE* fp=fopen((level_filename+".sol").c_str(), "wb");

    if(! fp) {
        return MessageBox(hwnd, "Couldn't save sollution.\r\n"
            "Are you sure you want to conitnue without saving?"
            " (you should answer 'no' and save your work manually)",
            "I/O Error",
            MB_YESNO|MB_ICONWARNING)==IDYES;
    }

    LRESULT res=SendMessage(edit, WM_GETTEXT, sizeof(program_buffer),
        (LPARAM)program_buffer);

    if(fwrite(program_buffer, 1, res, fp)!=(unsigned)res){
        fclose(fp);
        return MessageBox(hwnd, "There was a problem saving your sollution.\r\n"
            "Are you sure you want to conitnue without saving?"
            " (you should answer 'no' and save your work manually)",
            "I/O Error",
            MB_YESNO|MB_ICONWARNING)==IDYES;
    }
    fclose(fp);
    return true;
}

// Attemp to reload a previously saved sollution
void autoload(HWND hwnd)
{
    if(level_filename.empty()) return;

    FILE* fp=fopen((level_filename+".sol").c_str(), "rb");

    if(!fp) return;

    std::size_t read=fread(program_buffer, 1, sizeof(program_buffer)-1, fp);
    program_buffer[read]=0;

    SendMessage(edit, WM_SETTEXT, 0, (LPARAM)program_buffer);

    fclose(fp);
}

int lexer(int* p)
{
    if(program_buffer_it==program_buffer+sizeof(program_buffer))
        return EOF;
    char c=*program_buffer_it;

    program_buffer_it++;

    if(std::isdigit(c)){
        std::string buffer;
        buffer.push_back(c);

        while(program_buffer_it!=program_buffer+sizeof(program_buffer) &&
            std::isdigit(*program_buffer_it)){
            buffer.push_back(*program_buffer_it);
            program_buffer_it++;
        }
        sscanf(buffer.c_str(), "%d", p);
        return 256+buffer.size();
    }

    return c;
}
// Set the current score to a new value (and update UI)
void set_score(int sc)
{
    int const percent=(int)(((float)sc/(float)current_level.points)*100);
    current_dyn_level.points=sc;
    sprintf(program_buffer, "Score: %d/%d (%d%%)", sc,
        current_level.points,
        percent);
    SendMessage(score_label, WM_SETTEXT, 0, (LPARAM)program_buffer);
    SendMessage(score_progress, PBM_SETPOS, percent, 0);
}
// Set the program size to its current value  (will update the max bytes)
void set_program_size()
{
    GetWindowText(edit, program_buffer, sizeof(program_buffer));
    program_buffer_it=program_buffer;

    if(strlen(program_buffer)!=sizeof(program_buffer)-1 &&
        program_buffer[strlen(program_buffer)-1]!='\n'){
        strcat(program_buffer, "\n");
    }
    parser::get_char=lexer;
    std::list<function> pgm;
    parser::yyparse(&pgm);

    sprintf(program_buffer, "Program size: %d/%d",
        byte_size(pgm), current_level.max_program_size);
    SendMessage(status, WM_SETTEXT, 0, (LPARAM)program_buffer);
}


// Stop running the program
void stop_running(HWND hwnd)
{
    if(current_state==state::stopped) return;
    if(current_state==state::running){
        pause_running(hwnd);
    }

    KillTimer(hwnd, timer_id);
    SetWindowText(hwnd, "Herbert");
    current_state=state::stopped;

    // Set the edit area read only off
    SendMessage(edit, EM_SETREADONLY, false, 0);

    current_dyn_level.reset(current_level.robot_dir,
                        current_level.robot_x, current_level.robot_y);
    set_score(0);
    invalidate_playground(hwnd);
}


// Primitive to move the robot straight
void robot_straight()
{
    int new_x=current_dyn_level.robot_x;
    int new_y=current_dyn_level.robot_y;
    int const old_x=current_dyn_level.robot_x;
    int const old_y=current_dyn_level.robot_y;

    switch(current_dyn_level.robot_dir){
        case direction::up: new_y--; break;
        case direction::down: new_y++; break;
        case direction::left: new_x--; break;
        case direction::right: new_x++; break;
        default: break;
    }

    if(new_y<0 || new_y>=conf::playground_unit_count_y) return;
    if(new_x<0 || new_x>=conf::playground_unit_count_x) return;

    switch(current_level.cells[new_x][new_y].type){
        case level::cell::wall:
            return; // Moves in a wall
        case level::cell::white:
        {
            // Press a button
            if(!(current_dyn_level.cells[new_x][new_y]&dyn_level::cell_info::pressed)){
                current_dyn_level.cells[new_x][new_y]|=dyn_level::cell_info::pressed;
                current_dyn_level.nr_white_pressed++;

                int const new_score=conf::compute_score(current_level.points,
                    current_level.nr_white_buttons,
                    current_dyn_level.nr_white_pressed,
                    current_level.max_program_size,
                    byte_size(cpu->program()));

                if(new_score>current_dyn_level.points){
                    current_dyn_level.points=new_score;
                    set_score(new_score);
                }
            }

            break;
        } // level::cell::white
        case level::cell::gray:
            current_dyn_level.unpress_all();
            current_dyn_level.nr_white_pressed=0;
            break;
        case level::cell::dot:
            break;
    }

    switch(current_dyn_level.robot_dir){
        case direction::down:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_down;
        break;
        case direction::up:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_up;
        break;
        case direction::left:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_left;
        break;
        case direction::right:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_right;
        break;
    }
    current_dyn_level.robot_x=new_x;
    current_dyn_level.robot_y=new_y;


}

// Primitive to move the robot left (the 'l' instruction)
void robot_left()
{
    switch(current_dyn_level.robot_dir){
        case direction::left:
            current_dyn_level.robot_dir=direction::down; break;
        case direction::right:
            current_dyn_level.robot_dir=direction::up; break;
        case direction::up:
            current_dyn_level.robot_dir=direction::left; break;
        case direction::down:
            current_dyn_level.robot_dir=direction::right; break;
        default: break;
    }
}

// Primitive to move the robot right (the 'r' instruction)
void robot_right()
{
    switch(current_dyn_level.robot_dir){
        case direction::left:
            current_dyn_level.robot_dir=direction::up; break;
        case direction::right:
            current_dyn_level.robot_dir=direction::down; break;
        case direction::up:
            current_dyn_level.robot_dir=direction::right; break;
        case direction::down:
            current_dyn_level.robot_dir=direction::left; break;
        default: break;
    }
}

// Draw static level structures (buttons & walls)
void draw_level(HDC hdc, level const& lev)
{
    draw_grid(hdc);
    for(int y=0; y<conf::playground_unit_count_y; ++y)
    for(int x=0; x<conf::playground_unit_count_x; ++x){
        switch(lev.cells[x][y].type){
            case level::cell::white:
                draw_button(hdc, button::white, x, y); break;
            case level::cell::gray:
                draw_button(hdc, button::gray, x, y); break;
            case level::cell::wall:
                if(lev.cells[x][y].wall_dir&level::cell::left)
                    draw_wall(hdc, x, y, direction::left);
                if(lev.cells[x][y].wall_dir&level::cell::right)
                    draw_wall(hdc, x, y, direction::right);
                if(lev.cells[x][y].wall_dir&level::cell::up)
                    draw_wall(hdc, x, y, direction::up);
                if(lev.cells[x][y].wall_dir&level::cell::down)
                    draw_wall(hdc, x, y, direction::down);
                break;
            default: break;
        }// switch(lev.cells
    }
}

// Draw dynamic level structures (pressed buttons & trails)
void draw_dyn_level(HDC hdc, dyn_level const& dlev)
{
    for(int y=0; y<conf::playground_unit_count_y; ++y)
    for(int x=0; x<conf::playground_unit_count_x; ++x){
        if(dlev.cells[x][y]&dyn_level::cell_info::pressed)
            draw_button(hdc, button::white_pressed, x, y);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_up)
            draw_trail(hdc, x,y,direction::up);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_down)
            draw_trail(hdc, x,y,direction::down);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_left)
            draw_trail(hdc, x,y,direction::left);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_right)
            draw_trail(hdc, x,y,direction::right);
    }
    draw_robot(hdc, dlev.robot_dir, dlev.robot_x, dlev.robot_y);

}

void select_current_instruction()
{
    machine::address const& ip=cpu->ip();
    std::list<stem>::const_iterator it=ip.ins;
    function const& fun=*ip.fun;
    if(it==fun.stats.end()){
        --it;
    }
    SendMessage(edit, EM_SETSEL, it->loc.start, it->loc.end);
}

HWND create_toolbar(HWND parent)
{
    HIMAGELIST hlist=ImageList_Create(16, 16, ILC_MASK, 10, 0);
    ImageList_AddMasked(hlist, LoadBitmap(global::hinstance, "IDB_TOOLBAR"),
        RGB(0xff, 0 , 0xff));

    int tbstyle=TBSTYLE_FLAT|TBSTYLE_LIST|TBSTYLE_TOOLTIPS;

    HWND tool=CreateWindowEx(0,
        TOOLBARCLASSNAME, "",
        WS_CHILD|WS_VISIBLE|tbstyle,
        0, 0, 100, 10,
        parent,
        NULL, // menu
        global::hinstance, //hinstance
        NULL // options
        );

    SendMessage(tool, TB_SETIMAGELIST, 0, (LPARAM)hlist);
    SendMessage(tool, TB_SETEXTENDEDSTYLE, 0, TBSTYLE_EX_MIXEDBUTTONS);

    TBBUTTON  buttons[]={
        // Image, command, state, style
        {0,  IDM_PLAY, TBSTATE_ENABLED, BTNS_BUTTON},
        {1,  IDM_STOP, TBSTATE_ENABLED, BTNS_BUTTON},
        {-1, -1, TBSTATE_ENABLED, BTNS_SEP},
        {3,  IDM_SPEED_STEP, TBSTATE_ENABLED, BTNS_CHECK|BTNS_GROUP},
        {4,  IDM_SPEED_NORMAL, TBSTATE_ENABLED|TBSTATE_CHECKED, BTNS_CHECK|BTNS_GROUP},
        {5,  IDM_SPEED_FAST, TBSTATE_ENABLED, BTNS_CHECK|BTNS_GROUP},
        {-1, -1, TBSTATE_ENABLED, BTNS_SEP}
    };

    buttons[0].iString=0;
    buttons[1].iString=1;
    buttons[2].iString=2;
    buttons[3].iString=2;
    buttons[4].iString=3;
    buttons[5].iString=4;

    SendMessage(tool, TB_ADDSTRING, 0,
        (LPARAM)"Play\0Stop\0Step by step\0Normal Speed\0Fast forward\0");

    SendMessage(tool, TB_ADDBUTTONS, sizeof(buttons)/sizeof(*buttons), (LPARAM)buttons);
    SendMessage(tool, TB_AUTOSIZE, 0, 0);

    SetClassLong(toolbar, GCL_HBRBACKGROUND, (LONG)GetStockObject(BLACK_BRUSH));
    return tool;

}

char const* playground_class_name="PLAYGROUNDCLASSNAME";

int WINAPI WinMain(HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    PSTR cmd_line,
    int cmd_show)
{
	machine cpu;
	::cpu=&cpu;

    InitCommonControls();

    global::hinstance=hInstance;

    char const main_wnd_class_name[]="MAINWNDCLASSNAME";

    // Register main window's class
    WNDCLASS wnd_class;
    wnd_class.style=0;//CS_HREDRAW|CS_VREDRAW;
    wnd_class.lpfnWndProc=main_wnd_proc;
    wnd_class.cbClsExtra=wnd_class.cbWndExtra=0;
    wnd_class.hInstance=hInstance;
    wnd_class.hIcon=LoadIcon(hInstance, MAKEINTRESOURCE(IDI_HERBERT));
    wnd_class.hCursor=LoadCursor(NULL, IDC_HAND);
    wnd_class.lpszMenuName="MAINMENU";
    wnd_class.hbrBackground=(HBRUSH)(COLOR_APPWORKSPACE+1);//(HBRUSH)GetStockObject(WHITE_BRUSH);
    wnd_class.lpszClassName=main_wnd_class_name;

    if(! RegisterClass(&wnd_class)){
        MessageBox(NULL, "Couldn't register main window class",
            "Fatal Error", MB_ICONERROR);
        return 1;
    }

    // Register playground window's class
    WNDCLASS pwnd_class;
    pwnd_class.style=0;//CS_HREDRAW|CS_VREDRAW;
    pwnd_class.lpfnWndProc=playground_wnd_proc;
    pwnd_class.cbClsExtra=pwnd_class.cbWndExtra=0;
    pwnd_class.hInstance=hInstance;
    pwnd_class.hIcon=NULL;
    pwnd_class.hCursor=LoadCursor(NULL, IDC_HAND);
    pwnd_class.lpszMenuName=NULL;
    pwnd_class.hbrBackground=(HBRUSH)GetStockObject(WHITE_BRUSH);//(HBRUSH)GetStockObject(WHITE_BRUSH);
    pwnd_class.lpszClassName=playground_class_name;

    if(! RegisterClass(&pwnd_class)){
        MessageBox(NULL, "Couldn't register window class",
            "Fatal Error", MB_ICONERROR);
        return 1;
    }

    // Creating the main window
    HWND hwnd=CreateWindow(main_wnd_class_name,
        "Herbert",
        WS_SYSMENU|WS_MINIMIZEBOX|WS_THICKFRAME|WS_CLIPCHILDREN,
        CW_USEDEFAULT, CW_USEDEFAULT, zoom_factor*conf::playground_width,
            zoom_factor*conf::playground_height,
        NULL, //parent
        NULL, // menu handle
        hInstance,
        NULL);

    if(! hwnd){
        MessageBox(hwnd, "Couldn't create window", "Fatal Error",
            MB_ICONERROR);
        return 1;
    }

    ShowWindow(hwnd, cmd_show);
    UpdateWindow(hwnd);

    // Program main loop
    MSG msg;
    while(GetMessage(&msg, NULL, 0, 0)){
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    return LRESULT(msg.wParam);
}

LRESULT CALLBACK playground_wnd_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static HBITMAP back_bit;
    static HDC back_dc;

   switch(msg){
        case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC hdc=BeginPaint(hwnd, &ps);

            if(! back_bit){
                back_dc=CreateCompatibleDC(hdc);
                back_bit=CreateCompatibleBitmap(hdc,
                    zoom_factor*conf::playground_width, zoom_factor*conf::playground_height);
                DeleteObject(SelectObject(back_dc, back_bit));
            }

            draw_level(back_dc, current_level);
            draw_dyn_level(back_dc, current_dyn_level);

            if(zoom_factor==1)
                BitBlt(hdc, 0, 0, zoom_factor*conf::playground_width, zoom_factor*conf::playground_height,
                    back_dc, 0, 0, SRCCOPY);
            else
                StretchBlt(hdc, 0, 0, zoom_factor*conf::playground_width, zoom_factor*conf::playground_height,
                    back_dc, 0, 0, conf::playground_width, conf::playground_height, SRCCOPY);

            EndPaint(hwnd, &ps);
            return 0;
        } // WM_PAINT
   }
   return DefWindowProc(hwnd, msg, wparam, lparam);
}

LRESULT CALLBACK main_wnd_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static int toolbar_height;
    static int status_height;

    switch(msg){
        case WM_CREATE:
        {
            // compute the value of conf::score_label_size
            {
            HDC hdc=GetWindowDC(hwnd);
            HGDIOBJ back=SelectObject(hdc, GetStockObject(DEFAULT_GUI_FONT));
            TEXTMETRIC tm;
            GetTextMetrics(hdc, &tm);

            conf::score_label_size=tm.tmHeight+tm.tmHeight/3;

            SelectObject(hdc, back);
            ReleaseDC(hwnd, hdc);
            }

            toolbar=create_toolbar(hwnd);

            status=CreateWindowEx(0, STATUSCLASSNAME,
                (char const*)NULL,
                WS_CHILD|WS_VISIBLE,
                0, 0, 0, 0,
                hwnd,
                (HMENU)0,
                global::hinstance,
                NULL);

            playground=CreateWindow(playground_class_name,
                (char const*)NULL,
                WS_CHILD|WS_VISIBLE,
                0,conf::score_progress_size+conf::score_label_size,
                zoom_factor*conf::playground_width,
                zoom_factor*conf::playground_height,
                hwnd, (HMENU)IDC_PLAYGROUND, global::hinstance,
                NULL);

            RECT status_rect; GetWindowRect(status, &status_rect);
            status_height=status_rect.bottom - status_rect.top;

            RECT toolbar_rec; GetWindowRect(toolbar, &toolbar_rec);
            toolbar_height=toolbar_rec.bottom-toolbar_rec.top;

            // Try to obtain a client area of
            //  zoom_factor*conf::playground_width X zoom_factor*conf::playground_height
            RECT client; GetClientRect(hwnd, &client);
            const int client_w=client.right;
            const int client_h=client.bottom;

            CREATESTRUCT& cs=*(CREATESTRUCT*)lparam;

            MoveWindow(hwnd, cs.x, cs.y,
                cs.cx-client_w+zoom_factor*conf::playground_width,
                cs.cy-client_h+zoom_factor*conf::playground_height+
                conf::edit_height+conf::separator_size
                +toolbar_height+status_height
                +conf::score_progress_size+conf::score_label_size, true);

            edit=CreateWindow("EDIT", "f:ssslf\r\nf\r\n",
                WS_CHILD|WS_VISIBLE|ES_MULTILINE|WS_BORDER
                | ES_AUTOHSCROLL| ES_AUTOVSCROLL | WS_VSCROLL | ES_NOHIDESEL,
                0,
                conf::score_progress_size+conf::score_label_size+
                zoom_factor*conf::playground_height+conf::separator_size+toolbar_height,
                zoom_factor*conf::playground_width,
                conf::edit_height,
                hwnd,
                (HMENU)IDC_EDIT,
                global::hinstance,
                NULL);

            score_progress=CreateWindow(PROGRESS_CLASS,
                "", WS_CHILD|WS_VISIBLE|PBS_SMOOTH,
                0,conf::score_label_size,
                zoom_factor*conf::playground_width,
                conf::score_progress_size,
                hwnd,
                (HMENU)0,
                global::hinstance,
                NULL);

            score_label=CreateWindow("STATIC",
                "Score: 0/100 (%hh0)", WS_CHILD|WS_VISIBLE|SS_SUNKEN,
                0,0,
                zoom_factor*conf::playground_width,
                conf::score_label_size,
                hwnd,
                (HMENU)0,
                global::hinstance,
                NULL);

            SendMessage(score_progress, PBM_SETPOS, 50, 0);
            SendMessage(edit, WM_SETFONT, (WPARAM)GetStockObject(ANSI_FIXED_FONT),
                true);
            SendMessage(edit, EM_SETSEL, 0, -1);
            SetFocus(edit);

            SendMessage(hwnd, WM_COMMAND, MAKELPARAM(IDC_EDIT, EN_CHANGE),
                (LPARAM)edit);

            SendMessage(score_label, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT),
                true);


            SendMessage(hwnd, WM_COMMAND, MAKELPARAM(IDM_LEVEL_CLEAR,
                IDM_LEVEL_CLEAR), 0);

			CheckMenuRadioItem(
                        GetSubMenu(GetMenu(hwnd), 1),
                        IDM_ZOOM_100,
                        IDM_ZOOM_300,
                        IDM_ZOOM_100,
                        MF_BYCOMMAND);

            return 0;
        }

        case WM_ERASEBKGND:
        {
            HDC hdc=(HDC)wparam;
            RECT re; GetClientRect(hwnd, &re);
            RECT re2={0, conf::score_progress_size+conf::score_label_size,
                re.right-re.left,
                zoom_factor*conf::playground_height+conf::score_progress_size+
                conf::score_label_size};

            FillRect(hdc, &re, (HBRUSH)GetSysColorBrush(COLOR_3DFACE));
            FillRect(hdc, &re2, (HBRUSH)GetSysColorBrush(COLOR_APPWORKSPACE));
            return true;
        }
        case WM_SIZING:
        {
            RECT& new_rect=*(RECT*)lparam;
            int const new_w=new_rect.right-new_rect.left;
            int const new_h=new_rect.bottom-new_rect.top;

            int const min_w=zoom_factor*conf::playground_width;
            int const min_h=zoom_factor*conf::playground_height+status_height+
                toolbar_height+conf::edit_height+conf::separator_size+
                conf::score_progress_size+conf::score_label_size;

            if(new_w >= min_w && new_h >= min_h) return 0;
            // else restrict sizing

            switch(wparam){
                default:
                case WMSZ_BOTTOM:
                case WMSZ_BOTTOMRIGHT:
                case WMSZ_RIGHT:
                    new_rect.right=new_rect.left+(std::max)(min_w, new_w);
                    new_rect.bottom=new_rect.top+(std::max)(min_h, new_h);
                    break;
                case WMSZ_TOP:
                case WMSZ_TOPLEFT:
                case WMSZ_LEFT:
                    new_rect.left=new_rect.right-(std::max)(min_w, new_w);
                    new_rect.top=new_rect.bottom-(std::max)(min_h, new_h);
                    break;
                case WMSZ_TOPRIGHT:
                    new_rect.right=new_rect.left+(std::max)(min_w, new_w);
                    new_rect.top=new_rect.bottom-(std::max)(min_h, new_h);
                    break;
                case WMSZ_BOTTOMLEFT:
                    new_rect.left=new_rect.right-(std::max)(min_w, new_w);
                    new_rect.bottom=new_rect.top+(std::max)(min_h, new_h);
                    break;
            }

            return true;
        }  // WM_SIZING
        case WM_SIZE:
        {
            int const width=LOWORD(lparam);
            int const height=HIWORD(lparam);

            SendMessage(status, WM_SIZE, 0, 0);
            SendMessage(toolbar, WM_SIZE, 0, 0);

            RECT re;GetWindowRect(toolbar, &re);
            toolbar_height=re.bottom-re.top;

            GetWindowRect(status, &re);
            status_height=re.bottom-re.top;

            MoveWindow(score_label, 0, 0,
                width, conf::score_label_size, true);

            MoveWindow(score_progress, 0, conf::score_label_size,
                width, conf::score_progress_size, true);

            MoveWindow(toolbar, 0, zoom_factor*conf::playground_height+
                conf::score_progress_size+conf::score_label_size,
                width, toolbar_height, true);

            MoveWindow(status, 0, height-status_height,
                width, status_height, true);

            MoveWindow(playground, (width-zoom_factor*conf::playground_width)/2,
                conf::score_progress_size+conf::score_label_size,
                zoom_factor*conf::playground_width, zoom_factor*conf::playground_height, true);

            MoveWindow(edit, 0, zoom_factor*conf::playground_height+toolbar_height+
                conf::separator_size+conf::score_progress_size+
                conf::score_label_size, width,
                height-zoom_factor*conf::playground_height-toolbar_height-
                conf::score_progress_size-conf::separator_size-status_height-
                conf::score_label_size,
                true);

            return 0;
        }// WM_SIZE

        case WM_CLOSE:
            if(! autosave(hwnd)) return 0;
            DestroyWindow(hwnd);
            return 0;

        case WM_DESTROY:
            PostQuitMessage(0);
            return 0;

        case WM_COMMAND:
        {
            switch(LOWORD(wparam)){
                case IDM_ZOOM_100:
                case IDM_ZOOM_200:
                case IDM_ZOOM_300:
                {
                    CheckMenuRadioItem(
                        GetSubMenu(GetMenu(hwnd), 1),
                        IDM_ZOOM_100,
                        IDM_ZOOM_300,
                        LOWORD(wparam),
                        MF_BYCOMMAND);

                    int const old_zoom=zoom_factor;

                    switch(LOWORD(wparam)){
                        case IDM_ZOOM_100: zoom_factor=1; break;
                        case IDM_ZOOM_200: zoom_factor=2; break;
                        case IDM_ZOOM_300: zoom_factor=3; break;
                    }
                    if(old_zoom==zoom_factor) return 0;

                    RECT re; GetWindowRect(hwnd, &re);
                    if (zoom_factor<old_zoom){
                        re.right-=conf::playground_width*(old_zoom-zoom_factor);
                        re.bottom-=conf::playground_height*(old_zoom-zoom_factor);
                    }

                    SendMessage(hwnd, WM_SIZING, (WPARAM)WMSZ_BOTTOMRIGHT,
                            (WPARAM)&re);
                    MoveWindow(hwnd, re.left, re.top, re.right-re.left,
                        re.bottom-re.top, true);

                    InvalidateRect(hwnd, NULL, true);
                    invalidate_playground(hwnd);
                    return 0;
                } // case IDM_ZOOM_300
                case IDC_EDIT:
                {
                    if(HIWORD(wparam)==EN_CHANGE)
                        set_program_size();
                    return 0;
                } // IDM_EDIT
                case IDM_LEVEL_OPEN:
                {
                    stop_running(hwnd);

                    if(! autosave(hwnd)) return 0;

                    OPENFILENAME ofn;
					ZeroMemory(&ofn, sizeof(ofn));

                    ofn.lStructSize=sizeof(OPENFILENAME);
                    ofn.hwndOwner=hwnd;
                    ofn.hInstance=global::hinstance;
                    ofn.lpstrFilter="Text files (*.txt)\0*.txt\0All files (*.*)\0*.*\0";
                    ofn.lpstrFile=program_buffer; program_buffer[0]=0;
					ofn.nFilterIndex=1;
                    ofn.nMaxFile=sizeof(program_buffer);
					ofn.Flags=OFN_EXPLORER;

                    if(GetOpenFileName(&ofn)){
						FILE* fp=fopen(program_buffer, "rb");
                        if(! fp){
                            MessageBox(hwnd, "Error opening", program_buffer, MB_ICONERROR);
                            return 0;
                        }
                        level_filename=program_buffer;
                        if(!read(&current_level, fp)){
                            MessageBox(hwnd, "Parse Error", program_buffer, MB_ICONERROR);
                        }
                        current_dyn_level.reset(current_level.robot_dir,
                                current_level.robot_x,
                                current_level.robot_y);
                        invalidate_playground(hwnd);
                        autoload(hwnd);
                        set_program_size();
                        set_score(0);
                    }

                    return 0;
                }// case IDM_LEVEL_OPEN
                case IDM_STOP:
                    stop_running(hwnd);
                    return 0;

                case IDM_LEVEL_CLEAR:
                    stop_running(hwnd);
                    current_level.reset();
                    current_dyn_level.reset();
                    current_level.nr_white_buttons=1;
                    set_score(0);
                    set_program_size();
                    invalidate_playground(hwnd);
                    level_filename.clear();

                    return 0;

                case IDM_PAUSE:
                {
                    pause_running(hwnd);
                    return 0;
                }// IDM_PAUSE

                case IDM_SPEED_FAST:
                {
                    current_speed=speed::fast;
                    return 0;
                }// IDM_SPEED_FAST;
                case IDM_SPEED_NORMAL:
                {
                    current_speed=speed::normal;
                    return 0;
                }// IDM_SPEED_FAST;
                case IDM_SPEED_STEP:
                {
                    if(current_state==state::running)
                        select_current_instruction();
                    current_speed=speed::step;
                    return 0;
                }// IDM_SPEED_FAST;
                case IDM_PLAY:
                {
                    if(current_state==state::paused){
                        start_running(hwnd);
                        return 0;
                    }

                    stop_running(hwnd);

                    GetWindowText(edit, program_buffer, sizeof(program_buffer));
                    program_buffer_it=program_buffer;

                    parser::get_char=lexer;
                    std::list<function> pgm;
                    if(! parser::yyparse(&pgm)) {
                        MessageBox(hwnd, "Syntax error",
                        "Parse Error", MB_ICONERROR);
                        if(parser::error_loc!=0) parser::error_loc--;
                        SendMessage(edit, EM_SETSEL, parser::error_loc,
                            parser::error_loc+1);
                        return 0;
                    }

                    cpu->load(pgm);

                    instruction_buffer.clear();
                    start_running(hwnd);

                    return 0;
                }
                case IDM_HELP_ABOUT:
                    about_box::show(hwnd);
                    return 0;
                case IDM_HELP_BUGS:
                    bug_box::show(hwnd);
                    return 0;

            }
            return 0;
        } // case WM_COMMAND
        case WM_TIMER:
        {
            if(current_level.nr_white_buttons<=current_dyn_level.nr_white_pressed){
                KillTimer(hwnd, timer_id);
                sprintf(program_buffer, "You got %d points out of %d",
                    current_dyn_level.points, current_level.points);

                MessageBox(hwnd, program_buffer,  "Congratulation!", MB_ICONINFORMATION);
                stop_running(hwnd);
                return 0;
            }

            if(instruction_buffer.empty()) {
                std::vector<char> res;
                try {
                    if(!cpu->exec(res)){
                        pause_running(hwnd);
                        return 0;
                    }
                } catch(char const* str){
                    stop_running(hwnd);
                    select_current_instruction();
                    MessageBox(hwnd, str, "Runtime Error", MB_ICONERROR);
                    return 0;
                } catch(...) {
                    stop_running(hwnd);
                    select_current_instruction();
                    MessageBox(hwnd, "Unexpected exception", "Runtime Error",
                        MB_ICONERROR);
                    return 0;
                }

                instruction_buffer.assign(res.begin(),
                    res.end());

            } // if(instruction_buffer.empty())

            if(! instruction_buffer.empty()){
                switch(instruction_buffer.front()){
                    case 's': robot_straight(); break;
                    case 'l': robot_left(); break;
                    case 'r': robot_right(); break;
                    default: break;
                }
                instruction_buffer.pop_front();
                invalidate_playground(hwnd);
            }
            switch(current_speed){
                case speed::step: pause_running(hwnd); break;
                case speed::normal: SetTimer(hwnd, timer_id, timer_interv_normal, 0); break;
                case speed::fast: SetTimer(hwnd, timer_id, timer_interv_fast , 0); break;
            }
            if(current_speed==speed::step && instruction_buffer.empty())
                    select_current_instruction();
            return 0;
        } // WM_TIMER
    }
    return DefWindowProc(hwnd, msg, wparam, lparam);
}
